/**
 * Created by IntelliJ IDEA.
 * User: Kilho Shin
 * Date: 11/02/21
 * Time: 11:44
 * To change this template use File | Settings | File Templates.

This file is part of Mapping Kernels for Trees.

Mapping Kernels for Trees is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Mapping Kernels for Trees is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Mapping Kernels for Trees.  If not, see <http://www.gnu.org/licenses/>.

For more details about this code, please look at the companion paper
"Mapping Kernels for Trees", K. Shin, M.Cuturi, T.Kuboyama, Proceedings of
the International Conference on Machine Learning 2011.
 */

import scala.actors._
import scala.actors.Actor._
import scala.collection.mutable.Map
import scala.collection.mutable.ListBuffer
import scala.io.Source
import java.io.FileOutputStream
import java.io.OutputStreamWriter
import scala.util.matching.Regex
import java.util.Date
import java.text.SimpleDateFormat

/*
 * Define messages to be exchanged between Actors.
 * Organizer is the actor that manages the following actors.
 * ParseManager:  Converts trees that are represented using braces into a set of Vertex objects.
 *                In addition, it generates RefNode objects that represent the substructures
 *                to be used in computation of the target Gram matrix.
 * DecompManager: Aligns the RefNode objects in the order of dependencies.
 * CompManager:   Computes the target Gram matrix.
 */

// Messages for common use
case class Start(any: Any)
case class Finish(actor: Actor)
case class Exit
case class Do(any: Any)
case class Return(actor: Actor, any: Any)

/*
// Messages for parsing
case class PrintTree(i:Int, s: String) // For test only
case class Parse(l: List[(Int, String)])
case class RgstTree(v: Vertex)
case class RgstFrst(tpl: (RefNode, List[Vertex]))
case class RgstFrstWithReply(tpl: (RefNode, List[Vertex]))

// Messages for decomposing
//case class StartDecomp(m: Map[Int, List[RefNode]])
case class Decompose(l: List[RefNode])
case class Add(p: Extractor, lb: ListBuffer[RefNode])


// Messages for computing
//case class StartComp(db: List[(Signature, RefNode)])
//case class FillRect(l1: List[RefNode], l2: List[RefNode], l3: List[RefNode])
case class SetUp(scope: List[RefNode])
case class CalcRectangle(target: List[RefNode])
case class CalcTriangle(target: List[RefNode])
case class RgstValue(l: List[(((RefNode, RefNode)=>Bn, RefNode, RefNode), Bn)])
case class Calc(xstr: List[RefNode], ystr: List[RefNode])
//case class FillTriangle
*/


class Organizer(
  val filename: String,
  var outfile: String,
  val limit: Int,
  val pn: Int,
  val eval: ((RefNode, RefNode)=>Bn, RefNode, RefNode)=>Bn,
  val eval_db: Map[(Signature, Signature), Map[((RefNode, RefNode)=>Bn, RefNode, RefNode), Bn]],
  val krnList: List[(RefNode, RefNode)=>Bn],
  val krnlName : Map[(RefNode, RefNode)=>Bn, String],
  val krnlWrite: List[(RefNode, RefNode)=>Bn],
  val cubic: Boolean,
  val logSwitch: Boolean,
  val statSwitch: Boolean,
  val debug: Boolean,
  val quiet: Boolean,
  val naive: Boolean) extends Actor {

  val name_ext = new Regex("""(.*)\.([a-zA-Z]*)""")
  val name_ext(n, e) = filename
  val strt = new Date()
  val date_format = new SimpleDateFormat("-yyyy'-'MM'-'dd'-'HH'-'mm")
  if(outfile == null)
    outfile = n+date_format.format(strt)
  val out = new OutputStreamWriter(new FileOutputStream(outfile+".kernel"), "utf-8")
  var log: Option[OutputStreamWriter] = None
  if(logSwitch) {
    log = Some(new OutputStreamWriter(new FileOutputStream(outfile+".log"), "utf-8"))
    log.get.write("Program started at " + date_format.format(strt)+"\r\n")
    log.get.write("Working on "+filename+".\r\n")
    log.get.write("To calculate the following "+krnList.size+" kernels\r\n")
    krnList.foldLeft(1){
      (x,y) =>
      log.get.write(x+". "+krnlName(y)+"\r\n")
      x+1
    }
    log.get.write("To output the following "+krnlWrite.size+" kernels\r\n")
    krnlWrite.foldLeft(1){
      (x,y) =>
      log.get.write(x+". "+krnlName(y)+"\r\n")
      x+1
    }
    log.get.write("Parallelism = "+pn+"\r\n")
  }

  var stat: OutputStreamWriter = null
  if(statSwitch)
    stat = new OutputStreamWriter(new FileOutputStream(outfile+".stat"), "utf-8")

  var refs: Map[Int, List[RefNode]] = Map()
  val db: Map[Int, Option[Vertex]] = Map() //永続
  var str: List[(Int, Set[RefNode])] = Nil //永続

  var blocks: List[(List[Int], List[(Int, Set[RefNode])])] = Nil



  def act {
    val strt = System.currentTimeMillis()
    loop {
      react {
        case Finish(p) =>
          val stop = System.currentTimeMillis()
          print("*** Calculating kernel values took ")
          print(((stop - strt) / 60000) + " min ")
          print(((stop - strt) / 1000 % 60) + " sec ")
          println(((stop - strt) % 1000) + " msec.")
          out.close()
          if(logSwitch) {
            log.get.write("Calculating kernel values took ")
            log.get.write(((stop - strt) / 60000) + " min ")
            log.get.write(((stop - strt) / 1000 % 60) + " sec ")
            log.get.write(((stop - strt) % 1000) + " msec.\r\n")
            log.get.close()
          }
          exit()
      }
    }
  }

  def prepare: Boolean = {
    val src: List[(Int, String)] = prepare_source_list
    var state = true
    /*
     * Parsing input trees
     */

    println("* Parsing trees.")

    val parser = new TParser(src, refs, db, cubic)

    val strt = System.currentTimeMillis()

    parser.parse

    println("*** "+db.size+" trees have been parsed.")

    db.foreach(tpl =>
      if(tpl._2 == None) {
        println("*** Found error at line "+(tpl._1+1))
        state = false
      }
    )

    if(state) {
      /*
       * Extracting structures
       */

      println("* Extracting structures.")

      str = extract_structures
      refs = null

      println("*** Parsing and structure extraction took " + ((System.currentTimeMillis() - strt)) + " msec.")
      if(logSwitch)
        log.get.write("Parsing and structure extraction took " + ((System.currentTimeMillis() - strt)) + " msec.\r\n")


      if(debug) print_extracted_structures

      val str_n: Int = str.foldLeft(0){_+_._2.size}
      //if(check)
      println("*** " + str_n + " structures have been successfully extracted.")

      if(statSwitch) {
        println("* Collecting statistic data.")
        output_statistics
      }

      println("* Partitioning the input trees.")

      blocks = find_optimal_partition

      println("*** Upper-limit of # of structures involved in each block is "+limit+".")
      println("*** Optimal partition consists of "+blocks.size+" blocks.")
      print("***** ")
      blocks.foreach{
        x =>
        print("["+x._1.min+".."+x._1.max+"](")
        print(x._2.foldLeft(0)(_+_._2.size)+") ")
      }
      println(".")
      if(logSwitch)
        print_blocks

      if(debug)
        show_detail_of_blocks
      // val block_size: Int = if(bn > 0) (db.size+bn-1)/bn else 1

      println("* Compting kernel values.")

      if(naive)
        write_kernel_values_in_naive_mode
      else {
        val dcpmgr = new DivCompManager(this, db, str, krnList, krnlName,
                                    eval, eval_db, pn, blocks, krnlWrite, out, log, debug, quiet)
        dcpmgr.start
      }
    }
    state
  }

  def write_kernel_values_in_naive_mode {
    println("\n******* Total number of kernel values precalculated = 0, This is Naive Mode.")
    println("***** Outputing")
    val strt = System.currentTimeMillis()
    for(k<-krnlWrite) {
      out.write("*** "+krnlName(k)+"]\r\n")
      for(i <- 0 until db.size){
        var s = i+": "
        for(j <- 0 until db.size) {
          val r1 = db(i).get.children.head.refAsTree
          val r2 = db(j).get.children.head.refAsTree
          var v: Bn = null
          s += j+":"+eval(k, r1, r2)+" "
          eval_db.keySet.foreach(eval_db.remove)
          System.gc
        }
        out.write(s+"\r\n")
      }
    }
    val stop = System.currentTimeMillis()
    print("*** Calculating kernel values took ")
    print(((stop - strt) / 60000) + " min ")
    print(((stop - strt) / 1000 % 60) + " sec ")
    println(((stop - strt) % 1000) + " msec.")
    out.close()
    if(logSwitch) {
      log.get.write("Calculating kernel values took ")
      log.get.write(((stop - strt) / 60000) + " min ")
      log.get.write(((stop - strt) / 1000 % 60) + " sec ")
      log.get.write(((stop - strt) % 1000) + " msec.\r\n")
      log.get.close()
    }
  }


  def prepare_source_list: List[(Int, String)] = {
    var srcList: List[String] = Source.fromFile(filename).getLines.toList //一時
    var temp: List[(Int, String)] = Nil
    val label_tree = new Regex("""(\S*)\s*:\s*(.*)""")
    out.write("*** Labels\r\n")
    var count: Int = 0
    srcList.foreach{line =>
      val label_tree(l, t) = line
      out.write(count+":"+l+"\r\n")
      temp ::= (count, t)
      count += 1
    }
    temp.reverse
  }

  def extract_structures: List[(Int, Set[RefNode])] = {
    val temp: Map[Int, ListBuffer[RefNode]] = Map()
    for(x <- refs; y <- x._2){
      extract_from_root(y).foreach[Any]{
        z =>
        if(temp.isDefinedAt(z.size))
          temp(z.size) += z
        else
          temp += z.size -> ListBuffer(z)
      }
    }
    temp.toList.map(x=>(x._1, x._2.toSet)).sort(_._1 < _._1)
  }

  def extract_from_root(x: RefNode): List[RefNode] = {
    var temp: List[RefNode] = List(x)
    x.next.foreach(x=>temp :::= extract_from_root(x))
    temp
  }

  def find_optimal_partition: List[(List[Int], List[(Int, Set[RefNode])])] = {
    var temp: List[(List[Int], List[(Int, Set[RefNode])])] = Nil
    var current: List[(Int, Set[RefNode])] = Nil
    var bottom: Int = 0
    var top: Int = 0
    var skip: Int = 0
    while(bottom < db.size) {
      current = Nil
      str.foreach{
        x =>
        val tempSet = x._2.filter(!_.referencedFrom.intersect(bottom.to(top).toSet).isEmpty)
        current ::= x._1->tempSet
      }
      //println("main "+current.foldLeft(0)(_+_._2.size)+" ["+bottom+","+top+"] skip="+skip)
      if(current.foldLeft(0)(_+_._2.size) > limit)
        go_backward_by_one_until_condition_satisfied
      else
        go_forward_by_one_while_condition_satisfied
      temp ::= bottom.to(top).toList->current.reverse
      print(top+" ")
      skip = (skip + top - bottom)/2
      bottom = top+1
      top = if(bottom+skip < db.size) bottom+skip else db.size-1
    }

    def go_backward_by_one_until_condition_satisfied {
      while(current.foldLeft(0)(_+_._2.size) > limit && bottom < top) {
        top -= 1
        current = Nil
        str.foreach{
          x =>
          val tempSet = x._2.filter(!_.referencedFrom.intersect(bottom.to(top).toSet).isEmpty)
          current ::= x._1->tempSet
        }
      //println("backward "+current.foldLeft(0)(_+_._2.size)+" ["+bottom+","+top+"] skip="+skip)
      }
    }

    def go_forward_by_one_while_condition_satisfied {
      var prev: List[(Int, Set[RefNode])] = Nil
      while(current.foldLeft(0)(_+_._2.size) <= limit && top != db.size -1){
        top += 1
        prev = current
        current = Nil
        str.foreach{
          x =>
          val tempSet = x._2.filter(!_.referencedFrom.intersect(bottom.to(top).toSet).isEmpty)
          current ::= x._1->tempSet
        }
      //println("forward "+current.foldLeft(0)(_+_._2.size)+" ["+bottom+","+top+"] skip="+skip)
      }
      if(current.foldLeft(0)(_+_._2.size) > limit) {
        top -= 1
        current = prev
      }
    }
    println
    temp.reverse
  }

  def output_statistics {
    stat.write("Statistics for " + filename + "\r\n")
    stat.write("* # of input trees: " + db.size + "\r\n")
    stat.write("* # of structures extracted: " + str.foldLeft(0)(_+_._2.size) + "\r\n")

    var freq: Map[(Int, Int), Int] = Map()
    stat.write("* Histogram 1\r\n")
    stat.write("*** Row: Size of input trees\r\n")
    stat.write("*** Column: Height pf input trees\r\n")
    stat.write("*** Value: # of input trees \r\n")
    println("*** Histogram 1")

    db.foreach{x=>
      val root = x._2.get.children.head
      if(freq.isDefinedAt(root.size, root.height))
        freq(root.size, root.height) += 1
      else
        freq += (root.size, root.height)->1
    }
    val smax = freq.keySet.map(_._1).max
    val hmax = freq.keySet.map(_._2).max
    print_table(freq, 1.to(smax).toList, 1.to(hmax).toList)

    stat.write("* Histogram 2\r\n")
    stat.write("*** Row: Size of structures extracted\r\n")
    stat.write("*** Column: Height of structures extracted\r\n")
    stat.write("*** Value: # of structures extracted \r\n")
    println("*** Histogram 2")

    freq = Map()
    str.foreach(
      x=>1.to(hmax).foreach(h=>freq += (x._1, h)->x._2.filter(_.height==h).size)
      )
    print_table(freq, 1.to(smax).toList, 1.to(hmax).toList)

    stat.write("* Histogram 3\r\n")
    stat.write("*** Row: Size of structures extracted\r\n")
    stat.write("*** Column: Height of structures extracted\r\n")
    stat.write("*** Value: Total # of input trees including structures  \r\n")
    println("*** Histogram 3")

    freq = Map()
    for(x <- str; h <- 1 to hmax) {
      freq += (x._1, h)->x._2.filter(_.height==h).foldLeft(0)(_+_.referencedFrom.size)
    }
    print_table(freq, 1.to(smax).toList, 1.to(hmax).toList)

    var list: List[(Int, Int)] = Nil

    var skip: Int = 1
    if(db.size > 20)
      skip = db.size/20

    stat.write("* Histogram 4\r\n")
    stat.write("*** Row: Index of input trees\r\n")
    stat.write("*** Column: Size of structures extracted\r\n")
    stat.write("*** Value: # of structures in the input tree\r\n")
    println("*** Histogram 4")

    freq = Map()
    for(r <- 0 until db.size/skip; c <- 1 to smax) {
      freq += (r*skip, c)->str.find(_._1 == c).getOrElse((c, Set()))._2.filter(
        _.referencedFrom.contains(r*skip)).size
    }
    print_table(freq, 0.until(db.size/skip).toList.map(_*skip), 1.to(smax).toList)

    stat.write("* Histogram 5\r\n")
    stat.write("*** Row: Size of sets of input trees\r\n")
    stat.write("*** Column: Size of structures extracted\r\n")
    stat.write("*** Value: # of structures in the set of input trees\r\n")
    println("*** Histogram 5")

    freq = Map()
    for(r <- 1 to db.size/skip; c <- 1 to smax) {
      freq += (r*skip, c)->str.find(_._1 == c).getOrElse((c,Set()))._2.filter(
        !_.referencedFrom.intersect(0.until(r*skip).toSet).isEmpty).size
      /*
      if(str.isDefinedAt(c))
        freq += (r*skip, c)->str(c)._2.filter(!_.referencedFrom.intersect(0.until(r*skip).toSet).isEmpty).size
      else
        freq += (r*skip, c)->0
      */
    }
    print_table(freq, 1.until(db.size/skip).toList.map(_*skip), 1.to(smax).toList)

    stat.close
  }

    def print_table(t: Map[(Int, Int), Int], rLabels: List[Int], cLabels: List[Int]){
      val rmax = rLabels.size
      val cmax = cLabels.size
      print_sequence("*"::cLabels.map(_.toString))
      for(r <- rLabels){
        var line:List[Int] = List(r)
        for(c <- cLabels){
          if(t isDefinedAt (r,c))
            line ::= t(r,c)
          else
            line ::= 0
        }
        print_sequence(line.reverse.map(_.toString))
      }
    }

    def print_sequence(l: List[String]) {
      stat.write(l.head)
      l.tail.foreach(x=>stat.write(","+x))
      stat.write("\r\n")
    }

  def print_extracted_structures {
    str.foreach(_._2.foreach{x =>
        println(x.show)
        println("  Referenced From:"+x.referencedFrom.foldLeft(" ")(_+_+" "))
        print("  CUT: ")
        if(x.cut == null)
          print(" null ")
        else {
          val found = str.find(_._2.contains(x.cut))
          if(found == None)
            print("NO EXIST ERROR")
          else
            print(x.cut.show)
        }
        print("  TAKE: ")
        if(x.take == null)
          print(" null ")
        else {
          val found = str.find(_._2.contains(x.take))
          if(found == None)
            print("NO EXIST ERROR")
          else
            print(x.take.show)
        }
        print("  BRANCH: ")
        if(x.branch == null)
          print(" null ")
        else {
          val found = str.find(_._2.contains(x.branch))
          if(found == None)
            print("NO EXIST ERROR")
          else
            print(x.branch.show)
        }
        print("  CUT & BRANCH: ")
        if(x.cut_and_branch == null)
          print(" null ")
        else {
          val found = str.find(_._2.contains(x.cut_and_branch))
          if(found == None)
            print("NO EXIST ERROR")
          else
            print(x.cut_and_branch.show)
        }
        println
      }
    )
  }

  def print_blocks {
    log.get.write("The input trees are partitioned into the following "+blocks.size+" blocks.\r\n")
    blocks.foreach{
      x=>
      log.get.write("   ["+x._1.min+".."+x._1.max+"]: ")
      print_block_structure(x._2)
    }
  }

  def print_block_structure(x: List[(Int, Set[RefNode])]) {
      x.foreach{
        y =>
        log.get.write("("+y._1+","+y._2.size+") ")
      }
      log.get.write("Total = " + x.foldLeft(0)(_+_._2.size) + "\r\n")
  }

  def show_detail_of_blocks {
    blocks.foreach{
      x =>
      println("***** ["+x._1.min+".."+x._1.max+"]: ")
      x._2.foreach{
        y =>
        if(!y._2.isEmpty) {
          y._2.foreach{z=>print(z.show+", ")}
          println
        }
      }
    }
  }
}

