/**
 * Created by IntelliJ IDEA.
 * User: Kilho Shin
 * Date: 11/02/21
 * Time: 11:38
 * To change this template use File | Settings | File Templates.

This file is part of Mapping Kernels for Trees.

 Mapping Kernels for Trees is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Mapping Kernels for Trees is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Mapping Kernels for Trees.  If not, see <http://www.gnu.org/licenses/>.

 For more details about this code, please look at the companion paper
 "Mapping Kernels for Trees", K. Shin, M.Cuturi, T.Kuboyama, Proceedings of
 the International Conference on Machine Learning 2011.
 */

import java.util.Date
import scala.collection.mutable.Map
import java.text.SimpleDateFormat

object TreeKernel {

  /*
   * Definitions of kernel functions are provided.
   * eval(k, x, y) gives the kernel value k(x, y) for a substructure pair (x, y) of lower level.
   * cut: the remaining substructure by eliminating the leftmost subtree.
   * branch: the remaining substructure by eliminating the leftmost root
   */

   val eval_db: Map[(Signature, Signature), Map[((RefNode, RefNode)=>Bn, RefNode, RefNode), Bn]] = Map()

   val b0 = new Bn(0)
   val b1 = new Bn(1)
   val b2 = new Bn(2)
   val bErr = new Bn(-1000999)
   val X = new Bn(1,0,1.0)
   val Y = new Bn(0,1,1.0)

  def dirac(x: RefNode, y: RefNode): Bn = {
    if(x.label == y.label) b1 else b0
  }

  def cost(x: RefNode, y:RefNode): Bn = {
    if(x.label == y.label) X*X else X
  }

  def eval(f: (RefNode, RefNode)=>Bn, x: RefNode, y:RefNode): Bn = {
    var s: Signature = null
    var t: Signature = null

    def noExist: Bn = {
      // println("********** Value not found: kernel =" + krnlName(f))
      val value = f(x,y)
      if(t < s) {
        if(eval_db.isDefinedAt(t,s))
          eval_db(t, s) += (f,y,x) -> value
        else
          eval_db += (t,s) -> Map((f,y,x) -> value)
      } else {
        if(eval_db.isDefinedAt(s,t))
          eval_db(s, t) += (f,x,y) -> value
        else
          eval_db += (s,t) -> Map((f,x,y) -> value)
      }
      value
    }

    if(x == null || y == null)
      b0
    else {
      s = x.signature
      t = y.signature
      if(s == t) {
        if(eval_db isDefinedAt (s,t)) {
          if(eval_db(s,t) isDefinedAt (f,x,y))
            eval_db(s,t)(f,x,y)
          else if (eval_db(s,t) isDefinedAt (f,y,x))
            eval_db(s,t)(f,y,x)
          else
            noExist
        } else noExist
      } else if(s < t) {
        if(eval_db.isDefinedAt(s,t) && eval_db(s,t).isDefinedAt(f,x,y))
          eval_db(s,t)(f,x,y)
        else
          noExist
      } else if(t < s) {
        if(eval_db.isDefinedAt(t,s) && eval_db(t,s).isDefinedAt(f,y,x))
          eval_db(t,s)(f,y,x)
        else
          noExist
      } else {
        if(eval_db.isDefinedAt(s,t) && eval_db(s,t).isDefinedAt(f,x,y))
          eval_db(s,t)(f,x,y)
        else if(eval_db.isDefinedAt(t,s) && eval_db(t,s).isDefinedAt(f,y,x))
          eval_db(t,s)(f,y,x)
        else
          noExist
      }
    }
  }

  /*
  Count sparse forests.
  Complexity: Cubic.
  Dependency: sf.
   */
  val sf: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(sf, x.branch, y.branch)) * (b1 + eval(sf, x.cut, y.cut) ) +
        eval(sf, x.cut_and_branch, y) + eval(sf, x, y.cut_and_branch) -
        eval(sf, x.cut_and_branch, y.cut_and_branch)
  }

  /*
  Count sparse trees.
  Complexity Cubic.
  Dependency: st, sf.
   */
  val st: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(sf, x.branch, y.branch)) +
        eval(st, x.cut_and_branch, y) + eval(st, x, y.cut_and_branch) -
        eval(st, x.cut_and_branch, y.cut_and_branch)
//        eval(st, x.take, y.branch) + eval(st, x.branch, y.take) - eval(st, x.branch, y.branch) +
//        eval(st, x, y.cut) + eval(st, x.cut, y) - eval(st, x.cut, y.cut)
  }

  /*
  Count sparse trees rooted at the unique root of the parent tree.
  Constraints: Must not specify for kerneList..
  Complexity: Cubic.
  Dependency: sf.
   */
  val str: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(sf, x.branch, y.branch))
  }

  /*
  Count contiguous forests.
  Complexity: Cubic.
  Dependency: cf, cfr.
   */
  val cf: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cfr, x.branch, y.branch)) * (b1 + eval(cf, x.cut, y.cut)) +
        eval(cf, x.cut_and_branch, y) + eval(cf, x, y.cut_and_branch) -
        eval(cf, x.cut_and_branch, y.cut_and_branch)
  }

  /*
  Count contiguous forests rooted at roots of the parent forest.
  Each tree component of a substructure of this type is rooted at the root
  of a tree component of the parent forest.
  Constraints: Must not specify for kernelWrite.
  Complexity: Quadratic.
  Dependency: cfr.
   */
  val cfr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cfr, x.branch, y.branch)) * (b1 + eval(cfr, x.cut, y.cut)) +
        eval(cfr, x, y.cut) + eval(cfr, x.cut, y) - eval(cfr, x.cut, y.cut)
  }

  /*
  Count contiguous trees.
  Complexity: Quadratic.
  Dependency: ct, cfr.
   */
  val ct: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cfr, x.branch, y.branch)) +
        eval(ct, x.take, y.branch) + eval(ct, x.branch, y.take) - eval(ct, x.branch, y.branch) +
        eval(ct, x, y.cut) + eval(ct, x.cut, y) - eval(ct, x.cut, y.cut)
  }

  /*
  Count contiguous trees from the unique root of the parent tree
  assuming that the parent forest is a tree.  Quadratic.
  Constraints: Must not specify for kernList.
  Complexity: Quadratic.
  Dependency: cfr.
   */
  val ctr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      eval(cfr, x, y)
//      X * dirac(x, y) * (b1 + eval(cfr, x.branch, y.branch))
  }

  /*
  Count agreement trees.
  Complexity: Quadratic.
  Dependency: at, af.
   */
  val at: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(af, x.branch, y.branch)) +
        eval(at, x.take, y.branch) + eval(at, x.branch, y.take) - eval(at, x.branch, y.branch) +
        eval(at, x, y.cut) + eval(at, x.cut, y) - eval(at, x.cut, y.cut)
  }

  /*
  Count agreement forests.
  A different tree component of an agreement forest appears
  in a different tree component of its parent forest.
  Complexity: Quadratic.
  Dependency: af, at.
   */
    val af: (RefNode, RefNode) => Bn = {
      (x, y) =>
        (X * dirac(x, y) * (b1 + eval(af, x.branch, y.branch)) +
           eval(at, x.take, y.branch) + eval(at, x.branch, y.take) - eval(at, x.branch, y.branch)) *
             (b1 + eval(af, x.cut, y.cut)) +
          eval(af, x, y.cut) + eval(af, x.cut, y) - eval(af, x.cut, y.cut)
  }

  /*
  Count agreement trees rooted at the leftmost root.
  Constraints: Must not specify for kerneList.
  Complexity: Quadratic.
  Dependency: af
   */
  val atr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(af, x.branch, y.branch))
  }

  /*
  Count sparse paths.
  Complexity: Quadratic.
  Dependency: sp.
   */
  val sp: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(sp, x.branch, y.branch)) +
        eval(sp, x.take, y.branch) + eval(sp, x.branch, y.take) - eval(sp, x.branch, y.branch) +
        eval(sp, x, y.cut) + eval(sp, x.cut, y) - eval(sp, x.cut, y.cut)
  }

  /*
  Count sparse paths that are are rooted at the roots of the leftmost tree components.
  Constraints: Must not specify for kerneList.
  Complexity Quadratic.
  Dependency: sp.
   */
  val spr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(sp, x.branch, y.branch))
  }

  /*
  Count sparse path forests.
  A sparse path forest is a sequence of sparse paths such that different sparse paths in the sequence
  appear in different tree components.
  Hence, if a parent forest is a tree, a sparse path forest becomes a single sparse path.,
  Constraints: Must not specify for kernelWrite.
  Complexity: Quadratic.
  Dependency: spf, sp.
   */
  val spf: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(sp, x.branch, y.branch)) * (b1 + eval(spf, x.cut, y.cut)) +
        eval(spf, x.cut_and_branch, y) + eval(spf, x, y.cut_and_branch) -
        eval(spf, x.cut_and_branch, y.cut_and_branch)
  }

  /*
  Count sparse cascades.
  A sparse cascade is a sequence of sparse paths any two of which intersect each other
  only at the root of the leftmost tree component.
  Constraints: Must not specify for kerneList.
  Complexity: Quadratic.
  Dependency: sp, spf.
   */
  val scasc: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(spf, x.branch, y.branch))
  }

  /*
  Count contiguous paths.
  Complexity: Quadratic.
  Dependency: cp, cpr.
   */
  val cp: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cpr, x.branch, y.branch)) +
        eval(cp, x.take, y.branch) + eval(cp, x.branch, y.take) - eval(cp, x.branch, y.branch) +
        eval(cp, x, y.cut) + eval(cp, x.cut, y) - eval(cp, x.cut, y.cut)
  }

  /*
  Count contiguous paths rooted at the root of a tree component of the parent forest.
  Complexity: Quadratic.
  Dependency: cpr..
   */
  val cpr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cpr, x.branch, y.branch)) +
        eval(cpr, x.cut, y) + eval(cpr, x, y.cut) - eval(cpr, x.cut, y.cut)
  }

  /*
  Count contiguous path forests from roots.
  A contiguous path forest from root is a sequence of contiguous paths
  such that different paths in the sequence appear in different tree components and
  are rooted at the roots of their tree components.
  Constraints; Must not specify for kernelWrite.
  Complexity: Quadratic.
  Dependency: cpfr, cpr.
   */
  val cpfr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cpr, x.branch, y.branch)) * (b1 + eval(cpfr, x.cut, y.cut)) +
        eval(cpfr, x.cut, y) + eval(cpfr, x, y.cut) - eval(cpfr, x.cut, y.cut)
  }

  /*
  Count contiguous cascades.
  A contiguous cascade is a sequence of contiguous paths any two of which intersect each other
  only at the root of the leftmost tree component.
  Constraints: Must not specify for kerneList.
  Complexity: Quadratic.
  Dependency: cpr, cpfr.
   */
  val ccasc: (RefNode, RefNode) => Bn = { // Contiguous cascade from root
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cpfr, x.branch, y.branch))
  }

/*
   val cp: (RefNode, RefNode) => Bn = { // Contiguous path
    (x, y) =>
    X * dirac(x, y) * (b1 +
    x.children.foldLeft(b0)((s,t) => s+y.children.foldLeft(b0)(_+eval(cpr, t, _)))) +
    eval(cp, x.take, y.branch) + eval(cp, x.branch, y.take) - eval(cp, x.branch, y.branch) +
    eval(cp, x, y.cut) + eval(cp, x.cut, y) - eval(cp, x.cut, y.cut)
  }

  val cpr: (RefNode, RefNode) => Bn = { // Contiguous path from root
    (x, y) =>
    X * dirac(x, y) * (b1 +
    x.children.foldLeft(b0)((s,t) => s+y.children.foldLeft(b0)(_+eval(cpr, t, _))))
  }
*/

  /*
  Count co-rooted trees.
  Complexity: Quadratic.
  Dependency: cot, cofr.
   */
  val cot: (RefNode, RefNode) => Bn = {
    (x, y) =>
      X * dirac(x, y) * (b1 + eval(cofr, x.branch, y.branch)) +
        eval(cot, x.take, y.branch) + eval(cot, x.branch, y.take) - eval(cot, x.branch, y.branch) +
        eval(cot, x, y.cut) + eval(cot, x.cut, y) - eval(cot, x.cut, y.cut)
  }

  /*
  Count co-rooted forests from roots.
  A co-rooted forest from roots includes as many tree components as its parent forest.
  Furthermore, Each tree component of the co-rooted forest is rooted at
  the root of its parent tree component.
  Constraints: Must not specify for kernelWrite.
  Complexity: Quadratic.
  Dependency: cofr.
   */
  val cofr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      if(x.cut == null && y.cut == null)
        X * dirac(x, y) * (b1 + eval(cofr, x.branch, y.branch))
      else
        X * dirac(x, y) * (b1 + eval(cofr, x.branch, y.branch)) * eval(cofr, x.cut, y.cut)
  }

  /*
  Count co-rooted trees rooted at the unique root of the parent tree.
  Assume that the parent forest is a tree.
  Constraints: Must not specify for kerneList.
  Complexity: Cubic,
  Dependency: cofr.
   */
  val cotr: (RefNode, RefNode) => Bn = {
    (x, y) =>
      eval(cofr, x, y)
  }

  /*
  The tree kernel derived from Tai tree edit distance.
  Complexity: Cubic.
  Dependency: None.
   */
  val tai: (RefNode, RefNode) => Bn = {
    (x, y) =>
    cost(x, y) * (b1 + eval(tai, x.branch, y.branch)) * (b1 + eval(tai, x.cut, y.cut) ) +
    eval(tai, x.cut_and_branch, y) + eval(tai, x, y.cut_and_branch) -
    eval(tai, x.cut_and_branch, y.cut_and_branch)
  }
  /*
  The tree kernel derived from the constrained tree edit distance.
  Complexity: Quadratic
  Dependency: cd, cd_sub.
   */
  val cd: (RefNode, RefNode) => Bn = { // Constrained Distance
    (x, y) =>
    (cost(x, y) * (b1 + eval(cd, x.branch, y.branch)) +
      eval(cd, x.take, y.branch) + eval(cd, x.branch, y.take) - eval(cd, x.branch, y.branch)) *
    (eval(cd_sub, x.cut, y.cut) + b1) -
    eval(cd, x.branch, y.take) + eval(cd, x.branch, y) - eval(cd, x.branch, y.cut) -
    eval(cd, x.take, y.branch) + eval(cd, x, y.branch) - eval(cd, x.cut, y.branch) +
    eval(cd, x, y.cut) + eval(cd, x.cut, y) - eval(cd, x.cut, y.cut)
  }

  /*
  A tree kernel to be calculated to calculate cd.
  Constraints: Must not specify for kernelWrite.
  Complexity: Quadratic.
  Dependency: cd_sub, cd.
   */
  val cd_sub: (RefNode, RefNode) => Bn = {
    (x, y) =>
    (cost(x, y) * (b1 + eval(cd, x.branch, y.branch)) +
      eval(cd, x.take, y.branch) + eval(cd, x.branch, y.take) - eval(cd, x.branch, y.branch)) *
    (b1 + eval(cd_sub, x.cut, y.cut)) +
    eval(cd_sub, x, y.cut) + eval(cd_sub, x.cut, y) - eval(cd_sub, x.cut, y.cut)
  }

  /*
  The tree kernel derived from the more constrained tree edit distance.
  Complexity: Quadratic.
  Dependency: mcd, mcd_sub.
   */
    val mcd: (RefNode, RefNode) => Bn = { // More Constrained Distance
    (x, y) =>
    (cost(x, y) * (b1 + eval(mcd_sub, x.branch, y.branch)) +
      eval(mcd, x.take, y.branch) + eval(mcd, x.branch, y.take) - eval(mcd, x.branch, y.branch)) *
    (eval(mcd_sub, x.cut, y.cut) + b1) -
    eval(mcd, x.branch, y.take) + eval(mcd, x.branch, y) - eval(mcd, x.branch, y.cut) -
    eval(mcd, x.take, y.branch) + eval(mcd, x, y.branch) - eval(mcd, x.cut, y.branch) +
    eval(mcd, x, y.cut) + eval(mcd, x.cut, y) - eval(mcd, x.cut, y.cut)
  }

  /*
  A tree kernel necessary to calculate the mcd kernel.
  Constraints: Must not specify for kernelWrite.
  Complexity: Quadratic.
  Dependency: mcd_sub, mcd.
   */
  val mcd_sub: (RefNode, RefNode) => Bn = {
    (x, y) =>
    (cost(x, y) * (b1 + eval(mcd_sub, x.branch, y.branch)) +
      eval(mcd, x.take, y.branch) + eval(mcd, x.branch, y.take) - eval(mcd, x.branch, y.branch)) *
    (b1 + eval(mcd_sub, x.cut, y.cut)) +
    eval(mcd_sub, x, y.cut) + eval(mcd_sub, x.cut, y) - eval(mcd_sub, x.cut, y.cut)
  }

/*
  Sum of weighted sizes of congruent subforests
  Complexity: Cubic
  Dependency: sf
   */
  val size_sf: (RefNode, RefNode) => Bn = {
    (x, y) =>
    X * dirac(x, y) *
      ((b1 + eval(sf, x.branch, y.branch)) * eval(size_sf, x.cut, y.cut) +
        (b1 + eval(sf, x.branch, y.branch) + eval(size_sf, x.branch, y.branch)) *
          (b1 + eval(sf, x.cut, y.cut))) +
      eval(size_sf, x.cut_and_branch, y) + eval(size_sf, x, y.cut_and_branch) -
      eval(size_sf, x.cut_and_branch, y.cut_and_branch)
  }

/*
  Sum of weighted squared sizes of congruent subforests
  Complexity: Cubic
  Dependency: sf, size_sf
 */
  val sqsize_sf: (RefNode, RefNode) => Bn = {
    (x, y) =>
    X * dirac(x, y) *
    (eval(sqsize_sf, x.branch, y.branch) * (b1 + eval(sf, x.cut, y.cut)) +
      b2 * eval(size_sf, x.branch, y.branch) * eval(size_sf, x.cut, y.cut) +
      (b1 + eval(sf, x.branch, y.branch)) * eval(sqsize_sf, x.cut, y.cut) +
      b2 * (eval(size_sf, x.branch, y.branch) * (b1 + eval(sf, x.cut, y.cut)) +
        (b1 + eval(sf, x.branch, y.branch)) * eval(size_sf, x.cut, y.cut)) +
      (b1 + eval(sf, x.branch, y.branch)) * (b1 + eval(sf, x.cut, y.cut))) +
      eval(sqsize_sf, x.cut_and_branch, y) + eval(sqsize_sf, x, y.cut_and_branch) -
      eval(sqsize_sf, x.cut_and_branch, y.cut_and_branch)
  }

/*
  Partitionable kernels with hidden degree 2.and type II over sparse forests
  Complexity: Cubic
  Dependency: hd2_type2_sf_2
   */
  val hd2_type2_sf_1: (RefNode, RefNode) => Bn = {
    (x, y) =>
    dirac(x, y) *
      ((b1 + eval(hd2_type2_sf_1, x.branch, y.branch) + X * eval(hd2_type2_sf_2, x.branch, y.branch)) *
        (b1 + eval(hd2_type2_sf_1, x.cut, y.cut)) +
        X * (b1 + eval(hd2_type2_sf_1, x.branch, y.branch) + eval(hd2_type2_sf_2, x.branch, y.branch)) *
          eval(hd2_type2_sf_2, x.cut, y.cut)) +
    eval(hd2_type2_sf_1, x.cut_and_branch, y) + eval(hd2_type2_sf_1, x, y.cut_and_branch) -
    eval(hd2_type2_sf_1, x.cut_and_branch, y.cut_and_branch)
  }

  /*
  Partitionable kernels with hidden degree 2.and type II over sparse forests
  Complexity: Cubic
  Dependency: hd2_type2_sf_1
   */
  val hd2_type2_sf_2: (RefNode, RefNode) => Bn = {
    (x, y) =>
    dirac(x, y) *
      ((b1 + eval(hd2_type2_sf_1, x.branch, y.branch) + X * eval(hd2_type2_sf_2, x.branch, y.branch)) *
        eval(hd2_type2_sf_2, x.cut, y.cut) +
        (b1 + eval(hd2_type2_sf_1, x.branch, y.branch) + eval(hd2_type2_sf_2, x.branch, y.branch)) *
          (b1 + eval(hd2_type2_sf_1, x.cut, y.cut))) +
    eval(hd2_type2_sf_2, x.cut_and_branch, y) + eval(hd2_type2_sf_2, x, y.cut_and_branch) -
    eval(hd2_type2_sf_2, x.cut_and_branch, y.cut_and_branch)
  }

  /*
  Partitionable kernels with hidden degree 2.and type III over sparse forests
  Complexity: Cubic
  Dependency: hd2_type3_sf_2
   */
  val hd2_type3_sf_1: (RefNode, RefNode) => Bn = {
    (x, y) =>
    dirac(x, y) * 
      ((b1 + eval(hd2_type3_sf_1, x.branch, y.branch) + X * eval(hd2_type3_sf_2, x.branch, y.branch)) *
        (b1 + eval(hd2_type3_sf_1, x.cut, y.cut)) +
        X * (b1 + eval(hd2_type3_sf_1, x.branch, y.branch) + b2 * eval(hd2_type3_sf_2, x.branch, y.branch)) *
          eval(hd2_type3_sf_2, x.cut, y.cut)) + 
    eval(hd2_type3_sf_1, x.cut_and_branch, y) + eval(hd2_type3_sf_1, x, y.cut_and_branch) -
    eval(hd2_type3_sf_1, x.cut_and_branch, y.cut_and_branch)
  }

  /*
  Partitionable kernels with hidden degree 2.and type III over sparse forests
  Complexity: Cubic
  Dependency: hd2_type3_sf_1
   */
  val hd2_type3_sf_2: (RefNode, RefNode) => Bn = {
    (x, y) =>
    dirac(x, y) * 
      ((b1 + eval(hd2_type3_sf_1, x.branch, y.branch) + X * eval(hd2_type3_sf_2, x.branch, y.branch)) *
        eval(hd2_type3_sf_2, x.cut, y.cut) +
        (b1 + eval(hd2_type3_sf_1, x.branch, y.branch) + b2 * eval(hd2_type3_sf_2, x.branch, y.branch)) *
          (b1 + eval(hd2_type3_sf_1, x.cut, y.cut) + eval(hd2_type3_sf_2, x.cut, y.cut))) +
    eval(hd2_type3_sf_2, x.cut_and_branch, y) + eval(hd2_type3_sf_2, x, y.cut_and_branch) -
    eval(hd2_type3_sf_2, x.cut_and_branch, y.cut_and_branch)
  }
  
  var krnList: List[(RefNode, RefNode) => Bn] = Nil
  var krnlWrite: List[(RefNode, RefNode) => Bn] = Nil
  val krnlName: Map[(RefNode, RefNode)=>Bn, String] = Map(
    sf -> "count_sparse_forests",
    st -> "count_sparse_trees",
    str -> "count_sparse_tree_from_root",
    cf -> "count_contiguous_forests",
    ct -> "count_contiguous_trees",
    ctr -> "count_contiguous_trees_from_root",
    cfr -> "count_contiguous_forests_from_roots",
    at -> "count_agreement_trees",
    atr -> "count_agreement_trees_from_root",
    af -> "count_agreement_forests",
    sp -> "count_sparse_paths",
    spr -> "count_sparse_paths_from_root",
    scasc -> "count_sparse_cascades",
    spf -> "count_sparse_path_forests",
    cp -> "count_contiguous_paths",
    cpr -> "count_coutiguous_paths_from_root",
    ccasc -> "count_contiguous_cascades",
    cpfr -> "count_contiguous_path_forests",
    cot -> "count_corooted_trees",
    cotr -> "count_corooted_trees_from_root",
    cofr -> "count_corooted_forests",
    tai -> "tai_distance",
    cd -> "constrained_distance",
    cd_sub -> "constrained_distance_sub",
    mcd -> "more_constrained_distance",
    mcd_sub -> "more_constrained_distance_sub",
    size_sf -> "weighted_size_sparse_forests",
    sqsize_sf -> "weighted_squared_size_sparse_forests",
    hd2_type2_sf_1 -> "hidden_degree_2_type_2_diag_sf",
    hd2_type2_sf_2 -> "hidden_degree_2_type_2_compl_sf",
    hd2_type3_sf_1 -> "hidden_degree_2_type_3_diag_sf",
    hd2_type3_sf_2 -> "hidden_degree_2_type_3_compl_sf"
  )
  var cubic: Boolean = false

  /*
   * krnListに登録されたカーネルは登録順に評価される。
   */

  /**
   * @param args the command line arguments
   */


  def main(args: Array[String]): Unit = {
    var argIt = args.iterator
    var valid = true
    var filename: String = null
    var outfile: String = null
    var logSwitch: Boolean = false
    var stat: Boolean = false
    var debug: Boolean = false
    var quiet: Boolean = false
    var naive: Boolean = false
    var limit: Int = 1
    var parallel_n: Int = 1

    while(argIt.hasNext && valid) {
      argIt.next match{
        case "-f" =>
          if(argIt.hasNext)
            filename = argIt.next
          else valid = false
        case "-o" =>
          if(argIt.hasNext)
            outfile = argIt.next
          else valid = false
        case "-b" =>
          if(argIt.hasNext)
            limit = argIt.next.toInt
          else valid = false
        case "-p" =>
          if(argIt.hasNext)
            parallel_n = argIt.next.toInt
          else valid = false
        case "-all" =>
          krnList = List(sf, st, cf, ct, cfr, af, at, sp, spf,
            cp, cpr, cpfr, cot, cofr, tai, cd, cd_sub, mcd, mcd_sub)
          krnlWrite = List(sf, st, str, cf, ct, ctr, at, atr,
            sp, spr, scasc, cp, cpr, ccasc, cot, cotr, tai, cd, mcd, 
            hd2_type2_sf_1, hd2_type2_sf_2, hd2_type3_sf_1, hd2_type3_sf_2)
          cubic = true
        case "-cubic" =>
          krnList = List(sf, st, cfr, cf, sp, spf, tai)
          krnlWrite = List(sf, st, str, cf, scasc, tai)
          cubic = true
        case "-quad" =>
          krnList = List(ct, cfr, at, af, sp, cp, cpr, cpfr, cot, cot, cofr, cd, cd_sub, mcd, mcd_sub)
          krnlWrite = List(ct, ctr, at, atr, sp, spr, scasc, cp, cpr, cot, cotr, cd, mcd)
        case "-hd2/sf" =>
          krnList = List(hd2_type2_sf_1, hd2_type2_sf_2, hd2_type3_sf_1, hd2_type3_sf_2)
          krnlWrite = List(hd2_type2_sf_1, hd2_type2_sf_2, hd2_type3_sf_1, hd2_type3_sf_2)
          cubic = true
        case "-sf" =>
          krnList ::= sf
          krnlWrite ::= sf
          cubic = true
        case "-st" =>
          krnList :::= List(sf, st)
          krnlWrite ::= st
          cubic = true
        case "-str" =>
          krnList ::= sf
          krnlWrite ::= str
          cubic = true
        case "-cf" =>
          krnList :::= List(cfr, cf)
          krnlWrite ::= cf
          cubic = true
        case "-ct" =>
          krnList :::= List(cfr, ct)
          krnlWrite ::= ct
        case "-ctr" =>
          krnList ::= cfr
          krnlWrite ::= ctr
        case "-at" =>
          krnList :::= List(af, at)
          krnlWrite ::= at
        case "-atr" =>
          krnList :::= List(af, at)
          krnlWrite ::= atr
        case "-sp" =>
          krnList ::= sp
          krnlWrite ::= sp
        case "-spr" =>
          krnList ::= sp
          krnlWrite ::= spr
        case "-scasc" =>
          krnList :::= List(sp, spf)
          krnlWrite ::= scasc
          cubic = true
        case "-cp" =>
          krnList :::= List(cpr, cp)
          krnlWrite ::= cp
        case "-cpr" =>
          krnList ::= cpr
          krnlWrite ::= cpr
        case "-ccasc" =>
          krnList :::= List(cpr, cpfr)
          krnlWrite ::= ccasc
        case "-cot" =>
          krnList :::= List(cofr, cot)
          krnlWrite ::= cot
        case "-cotr" =>
          krnList ::= cofr
          krnlWrite ::= cotr
        case "-tai" =>
          krnList ::= tai
          krnlWrite ::= tai
          cubic = true
        case "-cd" =>
          krnList :::= List(cd_sub, cd)
          krnlWrite ::= cd
        case "-mcd" =>
          krnList :::= List(mcd_sub, mcd)
          krnlWrite ::= mcd
        case "-size/sf" =>
          krnList :::= List(sf, size_sf)
          krnlWrite :::= List(sf, size_sf)
          cubic = true
        case "-sqsize/sf" =>
          krnList :::= List(sf, size_sf, sqsize_sf)
          krnlWrite :::= List(sf, size_sf, sqsize_sf)
          cubic = true
        case "-hd2/type2/sf" =>
          krnList :::= List(hd2_type2_sf_1, hd2_type2_sf_2)
          krnlWrite :::= List(hd2_type2_sf_1, hd2_type2_sf_2)
          cubic = true
        case "-hd2/type3/sf" =>
          krnList :::= List(hd2_type3_sf_1, hd2_type3_sf_2)
          krnlWrite :::= List(hd2_type3_sf_1, hd2_type3_sf_2)
          cubic = true
        case "-log" =>
          logSwitch = true
        case "-stat" =>
          stat = true
        case "-debug" =>
          debug = true
        case "-quiet" =>                                          quiet = true
        case "-naive" =>
          naive = true
        case _ => valid = false
      }
    }
    if(filename == null) valid = false
    if(parallel_n < 1) valid = false
    if(krnList isEmpty) {
      krnList = List(sf, st, cf, ct, cfr, af, at, sp, spf,
            cp, cpr, cpfr, cot, cofr, tai, cd, cd_sub, mcd, mcd_sub,
            size_sf, sqsize_sf, hd2_type2_sf_1, hd2_type2_sf_2, hd2_type3_sf_1, hd2_type3_sf_2)
      krnlWrite = List(sf, st, str, cf, ct, ctr, at, atr,
            sp, spr, scasc, cp, cpr, ccasc, cot, cotr, tai, cd, mcd,
            size_sf, sqsize_sf, hd2_type2_sf_1, hd2_type2_sf_2, hd2_type3_sf_1, hd2_type3_sf_2)
      cubic = true
    }

    krnList = krnList.removeDuplicates
    krnlWrite = krnlWrite.removeDuplicates

    val start = new Date()
    val f = new SimpleDateFormat("' at' HH':'mm':'ss' on 'yyyy'/'MM'/'dd")
    println("Program started" + f.format(start)+"\r\n")
    println("* To calculate the following "+krnList.size+" kernels")
    krnList.foldLeft(1){
      (x,y) =>
      println(x+". "+krnlName(y))
      x+1
    }
    println("* To output the following "+krnlWrite.size+" kernels")
    krnlWrite.foldLeft(1){
      (x,y) =>
      println(x+". "+krnlName(y))
      x+1
    }
    println("* Prallelism: "+parallel_n)


    if(!valid)
      println("Invalid arguments.")
    else {
      val organizer = new Organizer(filename, outfile, limit, parallel_n,
                                eval, eval_db, krnList, krnlName, krnlWrite, cubic,
                                logSwitch, stat, debug, quiet, naive)
      if(organizer.prepare && !naive)
        organizer.start
    }
  }
}


